﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.IO;
using System.Linq;
using System.Text;
using System.Xml.Linq;
using DocumentFormat.OpenXml.Packaging;
using OpenXmlPowerTools;

namespace PlaceImage
{
    class Program
    {
        static void GetNormalFontInfo(SpreadsheetDocument sDoc, out string familyName, out string fontSize)
        {
            familyName = null;
            fontSize = null;

            WorkbookStylesPart stylesPart = sDoc
                .WorkbookPart
                .WorkbookStylesPart;
            XDocument stylesXDoc = stylesPart.GetXDocument();
            XElement normalCellStyle = stylesXDoc
                .Root
                .Elements(S.cellStyles)
                .Elements(S.cellStyle)
                .FirstOrDefault(cs => (string)cs.Attribute(SSNoNamespace.name) == "Normal");
            if (normalCellStyle != null)
            {
                string xfId = (string)normalCellStyle.Attribute(SSNoNamespace.xfId);
                XElement xf = stylesXDoc
                    .Root
                    .Elements(S.cellXfs)
                    .Elements(S.xf)
                    .FirstOrDefault(x => (string)x.Attribute(SSNoNamespace.xfId) == xfId);
                if (xf != null)
                {
                    int? fontId = (int?)xf.Attribute(SSNoNamespace.fontId);
                    if (fontId != null)
                    {
                        XElement font = stylesXDoc
                            .Root
                            .Elements(S.fonts)
                            .Elements(S.font)
                            .Skip((int)fontId)
                            .FirstOrDefault();
                        if (font != null)
                        {
                            familyName = (string)font
                                .Elements(S.name)
                                .Attributes(SSNoNamespace.val)
                                .FirstOrDefault();
                            fontSize = (string)font
                                .Elements(S.sz)
                                .Attributes(SSNoNamespace.val)
                                .FirstOrDefault();
                            return;
                        }
                    }
                }
            }
        }

        static int[] GetColumnWidthsInPixels(SpreadsheetDocument sDoc, WorksheetPart worksheet,
            int maxColumn)
        {
            string familyName;
            string fontSizeString;
            GetNormalFontInfo(sDoc, out familyName, out fontSizeString);

            float fontSize;
            fontSize = float.Parse(fontSizeString);
            Font normalFont = new Font(familyName, fontSize, GraphicsUnit.Point);

            // Need a Graphics object to measure font width
            float digitWidth = 0;
            Bitmap bm = new Bitmap(10, 10);
            bm.SetResolution(96, 96);

            Graphics g = Graphics.FromImage(bm);
            SizeF sf = g.MeasureString("0", normalFont, Int32.MaxValue,
                StringFormat.GenericTypographic);
            digitWidth = (int)sf.Width;
            g.Dispose();

            XDocument wsXDoc = worksheet.GetXDocument();

            List<float> charWidths = Enumerable.Range(1, maxColumn)
                .Select(c =>
                {
                    XElement colWidthElement = wsXDoc.Root
                        .Elements(S.cols)
                        .Elements(S.col)
                        .FirstOrDefault(z => c >= (int)z.Attribute(SSNoNamespace.min) &&
                            c <= (int)z.Attribute(SSNoNamespace.max));
                    if (colWidthElement != null)
                        return (float)colWidthElement.Attribute(SSNoNamespace.width);
                    else
                    {
                        float? defaultColumnWidth = (float?)wsXDoc
                            .Elements(S.sheetFormatPr)
                            .Attributes(SSNoNamespace.defaultColWidth)
                            .FirstOrDefault();
                        if (defaultColumnWidth != null)
                            return (float)defaultColumnWidth;
                        int unroundedPixelWidth = (int)(8 * digitWidth + 5);
                        int roundUpToMultipleOf64 = (unroundedPixelWidth + (8 - unroundedPixelWidth % 8));
                        float charWidth = roundUpToMultipleOf64 / digitWidth;
                        float charWidth2 = ((int)(charWidth * 256) / (float)(256));
                        return charWidth2;
                    }
                })
                .ToList();
            int[] pixelWidths = charWidths
                .Select(cw =>
                {
                    float widthInPixels = (float)Math.Truncate((double)(((cw * 256) +
                        (float)Math.Truncate((double)(128 / digitWidth))) / 256) * digitWidth);
                    return (int)widthInPixels;
                })
                .ToArray();
            return pixelWidths;
        }

        // 72 vertical points = 96 pixels
        static int[] GetRowHeightsInPixels(SpreadsheetDocument sDoc, WorksheetPart worksheet, 
            int maxRow)
        {
            XDocument wsXDoc = worksheet.GetXDocument();
            float? defaultRowHeight = (float?)wsXDoc
                .Root
                .Elements(S.sheetFormatPr)
                .Attributes(SSNoNamespace.defaultRowHeight)
                .FirstOrDefault();
            if (defaultRowHeight == null)
                defaultRowHeight = 15;
            Dictionary<int, float> customRowHeights = wsXDoc
                .Root
                .Elements(S.sheetData)
                .Elements(S.row)
                .Where(r => r.Attribute(SSNoNamespace.customHeight) != null ||
                    r.Attribute(SSNoNamespace.thickBot) != null ||
                    r.Attribute(SSNoNamespace.thickTop) != null)
                .ToDictionary(
                    r => (int)r.Attribute(SSNoNamespace.r),
                    r =>
                    {
                        float? h = (float?)r.Attribute(SSNoNamespace.ht);
                        if (h == null)
                            h = defaultRowHeight;
                        return (float)h;
                    });

            int[] rowHeightsInPixels = Enumerable.Range(1, maxRow)
                .Select(rowNum =>
                {
                    float rh = (float)defaultRowHeight;
                    if (customRowHeights.ContainsKey(rowNum))
                        rh = customRowHeights[rowNum];
                    int x = (int)(rh * 20 / 15);
                    return x;
                })
                .ToArray();
            return rowHeightsInPixels;
        }

        static void PlaceImageInSheet(SpreadsheetDocument sDoc,
            WorksheetPart worksheetPart,
            string imageFileName,
            int fromRow, int fromRowOffset,
            int fromCol, int fromColOffset,
            int toRow, int toRowOffset,
            int toCol, int toColOffset)
        {
            string rId = "R" + Guid.NewGuid().ToString().Replace("-", "");
            var drawingPart = worksheetPart.AddNewPart<DrawingsPart>("application/vnd.openxmlformats-officedocument.drawing+xml",
                rId);
            XDocument wsXDoc = worksheetPart.GetXDocument();
            if (wsXDoc.Root.Attribute(XNamespace.Xmlns + "x14ac") == null)
                wsXDoc.Root.Add(
                    new XAttribute(XNamespace.Xmlns + "x14ac", "http://schemas.microsoft.com/office/spreadsheetml/2009/9/ac"));
            if (wsXDoc.Root.Attribute(XNamespace.Xmlns + "r") == null)
                wsXDoc.Root.Add(
                    new XAttribute(XNamespace.Xmlns + "r", R.r.NamespaceName));

            XElement tablePart = wsXDoc.Descendants(S.tableParts).FirstOrDefault();
            if (tablePart != null)
                tablePart.AddBeforeSelf(
                    new XElement(S.drawing,
                        new XAttribute(R.id, rId)));
            else
                wsXDoc.Root.Add(
                    new XElement(S.drawing,
                        new XAttribute(R.id, rId)));

            worksheetPart.PutXDocument();

            string bmRid = "R" + Guid.NewGuid().ToString().Replace("-", "");
            var imagePart = drawingPart.AddNewPart<ImagePart>("image/png", bmRid);

            XDocument dXDoc = drawingPart.GetXDocument();
            string dxml = @"<xdr:wsDr xmlns:xdr='http://schemas.openxmlformats.org/drawingml/2006/spreadsheetDrawing' xmlns:a='http://schemas.openxmlformats.org/drawingml/2006/main'>
  <xdr:twoCellAnchor editAs='oneCell'>
    <xdr:from>
      <xdr:col>" + fromCol + @"</xdr:col>
      <xdr:colOff>" + fromColOffset + @"</xdr:colOff>
      <xdr:row>" + fromRow + @"</xdr:row>
      <xdr:rowOff>" + fromRowOffset + @"</xdr:rowOff>
    </xdr:from>
    <xdr:to>
      <xdr:col>" + toCol + @"</xdr:col>
      <xdr:colOff>" + toColOffset + @"</xdr:colOff>
      <xdr:row>" + toRow + @"</xdr:row>
      <xdr:rowOff>" + toRowOffset + @"</xdr:rowOff>
    </xdr:to>
    <xdr:pic>
      <xdr:nvPicPr>
        <xdr:cNvPr id='4' name='Picture 3'/>
        <xdr:cNvPicPr>
          <a:picLocks noChangeAspect='1'/>
        </xdr:cNvPicPr>
      </xdr:nvPicPr>
      <xdr:blipFill>
        <a:blip xmlns:r='http://schemas.openxmlformats.org/officeDocument/2006/relationships' r:embed='" +
                   bmRid + @"' cstate='print'>
          <a:extLst>
            <a:ext uri='{28A0092B-C50C-407E-A947-70E740481C1C}'>
              <a14:useLocalDpi xmlns:a14='http://schemas.microsoft.com/office/drawing/2010/main' val='0'/>
            </a:ext>
          </a:extLst>
        </a:blip>
        <a:stretch>
          <a:fillRect/>
        </a:stretch>
      </xdr:blipFill>
      <xdr:spPr>
        <a:xfrm>
          <a:off x='5054600' y='254000'/>
          <a:ext cx='1117600' cy='275945'/>
        </a:xfrm>
        <a:prstGeom prst='rect'>
          <a:avLst/>
        </a:prstGeom>
      </xdr:spPr>
    </xdr:pic>
    <xdr:clientData/>
  </xdr:twoCellAnchor>
</xdr:wsDr>";
            XElement droot = XElement.Parse(dxml);
            dXDoc.Add(droot);
            drawingPart.PutXDocument();

            byte[] byteArray = File.ReadAllBytes(imageFileName);

            using (MemoryStream ms = new MemoryStream(byteArray))
                imagePart.FeedData(ms);
        }

        static void Main(string[] args)
        {
            FileInfo fi = new FileInfo("../../Test2.xlsx");
            if (fi.Exists)
                fi.Delete();
            File.Copy("../../Test.xlsx", "../../Test2.xlsx");
            using (SpreadsheetDocument sDoc = 
                SpreadsheetDocument.Open("../../Test2.xlsx", true))
            {
                string sheet = "Sheet1";
                XDocument workbookXDoc = sDoc.WorkbookPart.GetXDocument();
                string sheet1Rid = (string)workbookXDoc
                    .Root
                    .Elements(S.sheets)
                    .Elements(S.sheet)
                    .Where(sh => (string)sh.Attribute(SSNoNamespace.name) == sheet)
                    .Attributes(R.id).FirstOrDefault();
                WorksheetPart sheetPart = (WorksheetPart)sDoc
                    .WorkbookPart
                    .GetPartById(sheet1Rid);
                if (sheetPart != null)
                {
                    Bitmap bmp = new Bitmap("../../PalaceGate.png");
                    SizeF d = bmp.PhysicalDimension;

                    //d.Width = (float)(d.Width * 1.6);
                    //d.Height = (float)(d.Height * 1.6);

                    int[] colWidths = GetColumnWidthsInPixels(sDoc, sheetPart, 20);
                    int[] colPositions = new [] { 0 }.Concat(colWidths.Rollup(0, (s, i) => s + i)).ToArray();

                    int[] rowHeights = GetRowHeightsInPixels(sDoc, sheetPart, 20);
                    int[] rowPositions = new[] { 0 }.Concat(rowHeights.Rollup(0, (s, i) => s + i)).ToArray();

                    for (int i = 0; i < colWidths.Length; ++i)
                        Console.WriteLine("Column {0}: {1} {2}", i + 1, colWidths[i], colPositions[i]);
                    for (int i = 0; i < rowHeights.Length; ++i)
                        Console.WriteLine("Row {0}: {1} {2}", i + 1, rowHeights[i], rowPositions[i]);

                    int startingColumn = 2;
                    int startingRow = 4;

                    int startingColumnPixelPosition = colPositions[startingColumn];
                    int endingColumnPixelPosition = startingColumnPixelPosition + (int)d.Width;
                    int endingColumn = colPositions.Select((cp, i) => new {
                        Position = cp,
                        Index = i,
                    }).LastOrDefault(cp => cp.Position <= endingColumnPixelPosition).Index;
                    int columnPixelOverlap = endingColumnPixelPosition - colPositions[endingColumn];
                    int colEmuOffset = (int)(((float)columnPixelOverlap / 96.0) * 914400.0);

                    int startingRowPixelPosition = rowPositions[startingRow];
                    int endingRowPixelPosition = startingRowPixelPosition + (int)d.Height;
                    int endingRow = rowPositions.Select((rp, i) => new
                    {
                        Position = rp,
                        Index = i,
                    }).LastOrDefault(rp => rp.Position <= endingRowPixelPosition).Index;
                    int rowPixelOverlap = endingRowPixelPosition - rowPositions[endingRow];
                    int rowEmuOffset = (int)(((float)rowPixelOverlap / 96.0) * 914400.0);


                    PlaceImageInSheet(sDoc,
                        sheetPart,
                        "../../PalaceGate.png", startingRow, 0, startingColumn, 0, endingRow, rowEmuOffset, endingColumn, colEmuOffset);
                }
            }
        }
    }
}
